@@ -37,6 +37,7 @@ Spatial calculations
 .. autosummary::
     kkpy.util.nn_idx_2d
     kkpy.util.cross_section_2d
+    kkpy.util.to_lower_reolution
     
 ICE-POP 2018
 --------------
@@ -51,9 +52,10 @@ Miscellaneous
     kkpy.util.std2d
     kkpy.util.nanstd2d
     kkpy.util.nanconvolve2d
+    kkpy.util.derivative
     kkpy.util.stats
     kkpy.util.summary
-"""                                                   
+"""
 import numpy as np
 
 def wind2uv(wd=None, ws=None, radians=False, knots=False):
@@ -70,9 +72,9 @@ def wind2uv(wd=None, ws=None, radians=False, knots=False):
         Array containing wind direction in **degree**. It should be **meteorological** direction, not mathematical.
     ws : array_like
         Array containing wind speed in **m/s**.
-    radians : bool, optional
+    radians : boolean, optional
         If this is set to True, the unit of *wd* is **radian**. The default is False (i.e. **degree**).
-    knots : bool, optional
+    knots : boolean, optional
         If this is set to True, the unit of *ws* is **knots**. The default is False (i.e. **m/s**).
         
     Returns
@@ -86,7 +88,7 @@ def wind2uv(wd=None, ws=None, radians=False, knots=False):
         wd = np.radians(wd)
     
     if knots:
-        ws = ms2knot(ws)
+        ws = knot2ms(ws)
     
     u = -ws * np.sin(wd)
     v = -ws * np.cos(wd)
@@ -258,7 +260,7 @@ def cross_section_2d(xy0, xy1, x2d, y2d, value2d, linewidth=1, along='x', reduce
     
     # get xaxis
     ixlen, iylen = iend - istart
-    ixsize = np.int(np.ceil(np.hypot(ixlen, iylen) + 1))
+    ixsize = int(np.ceil(np.hypot(ixlen, iylen) + 1))
     if along in 'x':
         xaxis = np.linspace(xy0[0], xy1[0], ixsize)
     else:
@@ -353,7 +355,7 @@ def dist_bearing(lonlat0, lonlat1, radians=False):
         Array containing longitude and latitude of the first point. Longitude (latitude) should be at the first (second) element.
     lonlat1 : 1D Array
         Array containing longitude and latitude of the second point. Longitude (latitude) should be at the first (second) element.
-    radians : bool, optional
+    radians : boolean, optional
         If this is set to True, the unit of *bearing* is **radian**. The default is False (i.e. **degree**).
         
     Returns
@@ -693,7 +695,7 @@ def icepop_sites():
         (128.805847, 37.738157, 175, 'BKC'),
         (128.758636, 37.686953, 855, 'CPO'),
         (128.718825, 37.677331, 773, 'DGW'),
-        (128.665208, 37.665208, 789, 'MHS'),
+        (128.699611, 37.665208, 789, 'MHS'),
         (128.670494, 37.643342, 772, 'YPO'),
         (128.564700, 38.250900,  18, 'SCW'),
         (128.629700, 38.087200,   4, 'YYO'),
@@ -791,7 +793,7 @@ def stats(x, y,
           fmtbias='.3f', fmtrmse='.3f',
           fmtstd='.3f', fmtcorr='.3f'):
     """
-    Get perpormance matrix BIAS, RMSE, STD, and CORR values and its f-string.
+    Get performance matrix BIAS, RMSE, STD, and CORR values and its f-string.
     
     Examples
     ---------
@@ -856,6 +858,7 @@ def summary(arr, cnt_print=10):
         print(f'First {cnt_print} values:\n {arr[:cnt_print]}\n')
         print(f'Last {cnt_print} values:\n {arr[-cnt_print:]}\n')
         
+        arr = arr[~np.isnan(arr)]
         arr_sorted = np.sort(arr)
         print(f'Lowest {cnt_print} values:\n {arr_sorted[:cnt_print]}\n')
         print(f'Highest {cnt_print} values:\n {arr_sorted[-cnt_print:]}\n')
@@ -869,3 +872,142 @@ def summary(arr, cnt_print=10):
             print(f'{x:.3f} -- {y:.0f}')
         
     return
+
+def derivative(var, cnt_x, pixelsize=1, axis=0):
+    """
+    Get derivative of 1D or 2D array.
+    
+    Examples
+    ---------
+    >>> # Derivative of 1D array with a window size of 5.
+    >>> # Note that the horizontal resolution of xarr is 0.1.
+    >>> xarr = np.arange(100)*0.1
+    >>> yarr = np.exp(xarr)
+    >>> dydx = kkpy.util.derivative(yarr, 5, pixelsize=0.1)
+    
+    >>> # This example gives the derivative of arr2d along x axis with a window size of 20.
+    >>> # Assume that you have an arr2d[iy,ix] with horizontal resolution of 3 km.
+    >>> dzdx = kkpy.util.derivative(arr2d, 20, pixelsize=3000, axis=1)
+    
+    Parameters
+    ----------
+    var : array_like
+        Array containing multiple variables and observations.
+    cnt_x : int
+        The number of the pixel for derivative. This should be odd number. The derivative will be (Y[i+N/2]-Y[i-N/2]) / (N-1).
+    pixelsize : float, optional
+        The size of one pixel of x axis. If default (1), the derivative means the rate of change on an interval of **one pixel** with resolution of 1.
+    axis : int, optional
+        Axis along which the derivatives are computed. Default is 0.
+        
+    Returns
+    ---------
+    derivative : array_like
+        Return a derivative for a given array.
+    """
+    import astropy.convolution
+
+    if cnt_x % 2 != 1:
+        raise ValueError('cnt_x must be odd.')
+    
+    if var.ndim > 2:
+        raise ValueError('dimension size of var should be <= 2')
+
+    kernel = np.zeros(cnt_x)
+    kernel[0] = 1
+    kernel[-1] = -1
+    
+    if var.ndim == 2:
+        if axis == 0:
+            kernel = kernel[:,np.newaxis]
+        if axis == 1:
+            kernel = kernel[np.newaxis,:]
+    
+    result = astropy.convolution.convolve(
+            var,
+            kernel,
+            boundary='extend',
+            normalize_kernel=False,
+            nan_treatment='fill',
+            fill_value=np.nan,
+            preserve_nan=True
+        ) / (cnt_x-1)
+    
+    if var.ndim == 1:
+        for j in np.arange(int(cnt_x/2)+1):
+            result[j] = (var[j+int(cnt_x/2)]-var[0]) / (j+int(cnt_x/2))
+        for j in np.arange(var.size-1-int(cnt_x/2), var.size):
+            result[j] = (var[-1]-var[j-int(cnt_x/2)]) / (var.size-j+int(cnt_x/2)-1)
+    
+    if var.ndim == 2:
+        if axis == 0:
+            for j in np.arange(int(cnt_x/2)+1):
+                result[j,:] = (var[j+int(cnt_x/2),:]-var[0,:]) / (j+int(cnt_x/2))
+            for j in np.arange(var.shape[0]-1-int(cnt_x/2), var.shape[0]):
+                result[j,:] = (var[-1,:]-var[j-int(cnt_x/2),:]) / (var.shape[0]-j+int(cnt_x/2)-1)
+        if axis == 1:
+            for j in np.arange(int(cnt_x/2)+1):
+                result[:,j] = (var[:,j+int(cnt_x/2)]-var[:,0]) / (j+int(cnt_x/2))
+            for j in np.arange(var.shape[1]-1-int(cnt_x/2), var.shape[1]):
+                result[:,j] = (var[:,-1]-var[:,j-int(cnt_x/2)]) / (var.shape[1]-j+int(cnt_x/2)-1)
+    
+    derivative = result / pixelsize
+    
+    return derivative
+
+def to_lower_resolution(arr_in, ratio_x, ratio_y, dBZ=False):
+    """
+    Get 2D array with lower resolution.
+    
+    Examples
+    ---------
+    >>> arr1 = np.random.rand(250,150) # shape: (250,150)
+    >>> arr2 = to_lower_resolution(arr1,10,10) # shape: (25,15)
+    >>> arr3 = to_lower_resolution(arr1,5,10) # shape: (50,15)
+    
+    Parameters
+    ----------
+    arr_in : 2D array
+        Array you want to lower the resolution.
+    ratio_x : int
+        The reduction ratio over x axis. The size of **arr_in** xaxis should be able to be divided by **ratio_x**.
+    ratio_y : int
+        The reduction ratio over y axis. The size of **arr_in** yaxis should be able to be divided by **ratio_y**.
+    dBZ : boolean, optional
+        True if return a linear-scale average of reflectivity (dBZ).
+        
+    Returns
+    ---------
+    arr_out : 2D array
+        Return an array with lower resolution.
+    
+    Notes
+    ---------
+    This code is from Stack Overflow answer (https://stackoverflow.com/a/14916963/12272819), written by Jaime (https://stackoverflow.com/users/110026/jaime).
+    This is licensed under the Creative Commons Attribution-ShareAlike 3.0 license (CC BY-SA 3.0).
+    """
+    
+    if arr_in.ndim != 2:
+        raise ValueError("arr_in must be a 2D array")
+    if ratio_x <= 1 or ratio_y <= 1:
+        raise ValueError("ratio_x and ratio_y must be greater than 1")
+    if not isinstance(ratio_x, int) or not isinstance(ratio_y, int):
+        raise ValueError("ratio_x and ratio_y must be an integer")
+    
+    nx_in, ny_in = arr_in.shape
+    nx_out, ny_out = nx_in/ratio_x, ny_in/ratio_y
+    
+    if not nx_out.is_integer():
+        raise ValueError(f"check ratio_x: nx_out={nx_out}")
+    if not ny_out.is_integer():
+        raise ValueError(f"check ratio_y: ny_out={ny_out}")
+    
+    arr_out = arr_in.reshape(
+        [int(nx_out), ratio_x, int(ny_out), ratio_y])
+    if not dBZ:
+        arr_out = arr_out.mean(axis=3).mean(axis=1)
+    else:
+        arr_out = kkpy.util.dbzmean(arr_out, axis=3)
+        arr_out = kkpy.util.dbzmean(arr_out, axis=1)
+    
+    return arr_out
